#
# Autogenerated by Thrift Compiler (0.15.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:package_prefix=qutrunk.sim.qusprout.
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import qutrunk.sim.qusprout.code.ttypes

from thrift.transport import TTransport
all_structs = []


class ExecCmdType(object):
    ExecTypeDefault = 0
    ExecTypeCpuSingle = 1
    ExecTypeCpuMpi = 2
    ExecTypeGpuSingle = 3

    _VALUES_TO_NAMES = {
        0: "ExecTypeDefault",
        1: "ExecTypeCpuSingle",
        2: "ExecTypeCpuMpi",
        3: "ExecTypeGpuSingle",
    }

    _NAMES_TO_VALUES = {
        "ExecTypeDefault": 0,
        "ExecTypeCpuSingle": 1,
        "ExecTypeCpuMpi": 2,
        "ExecTypeGpuSingle": 3,
    }


class PauliOperType(object):
    POT_PAULI_I = 0
    POT_PAULI_X = 1
    POT_PAULI_Y = 2
    POT_PAULI_Z = 3

    _VALUES_TO_NAMES = {
        0: "POT_PAULI_I",
        1: "POT_PAULI_X",
        2: "POT_PAULI_Y",
        3: "POT_PAULI_Z",
    }

    _NAMES_TO_VALUES = {
        "POT_PAULI_I": 0,
        "POT_PAULI_X": 1,
        "POT_PAULI_Y": 2,
        "POT_PAULI_Z": 3,
    }


class Amplitude(object):
    """
    Attributes:
     - reals
     - imags
     - startind
     - numamps

    """


    def __init__(self, reals=None, imags=None, startind=None, numamps=None,):
        self.reals = reals
        self.imags = imags
        self.startind = startind
        self.numamps = numamps

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.reals = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = iprot.readDouble()
                        self.reals.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.imags = []
                    (_etype9, _size6) = iprot.readListBegin()
                    for _i10 in range(_size6):
                        _elem11 = iprot.readDouble()
                        self.imags.append(_elem11)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.startind = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I32:
                    self.numamps = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Amplitude')
        if self.reals is not None:
            oprot.writeFieldBegin('reals', TType.LIST, 1)
            oprot.writeListBegin(TType.DOUBLE, len(self.reals))
            for iter12 in self.reals:
                oprot.writeDouble(iter12)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.imags is not None:
            oprot.writeFieldBegin('imags', TType.LIST, 2)
            oprot.writeListBegin(TType.DOUBLE, len(self.imags))
            for iter13 in self.imags:
                oprot.writeDouble(iter13)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.startind is not None:
            oprot.writeFieldBegin('startind', TType.I32, 3)
            oprot.writeI32(self.startind)
            oprot.writeFieldEnd()
        if self.numamps is not None:
            oprot.writeFieldBegin('numamps', TType.I32, 4)
            oprot.writeI32(self.numamps)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.reals is None:
            raise TProtocolException(message='Required field reals is unset!')
        if self.imags is None:
            raise TProtocolException(message='Required field imags is unset!')
        if self.startind is None:
            raise TProtocolException(message='Required field startind is unset!')
        if self.numamps is None:
            raise TProtocolException(message='Required field numamps is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Matrix(object):
    """
    Attributes:
     - reals
     - imags
     - unitary

    """


    def __init__(self, reals=None, imags=None, unitary=None,):
        self.reals = reals
        self.imags = imags
        self.unitary = unitary

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.reals = []
                    (_etype17, _size14) = iprot.readListBegin()
                    for _i18 in range(_size14):
                        _elem19 = []
                        (_etype23, _size20) = iprot.readListBegin()
                        for _i24 in range(_size20):
                            _elem25 = iprot.readDouble()
                            _elem19.append(_elem25)
                        iprot.readListEnd()
                        self.reals.append(_elem19)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.imags = []
                    (_etype29, _size26) = iprot.readListBegin()
                    for _i30 in range(_size26):
                        _elem31 = []
                        (_etype35, _size32) = iprot.readListBegin()
                        for _i36 in range(_size32):
                            _elem37 = iprot.readDouble()
                            _elem31.append(_elem37)
                        iprot.readListEnd()
                        self.imags.append(_elem31)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.BOOL:
                    self.unitary = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Matrix')
        if self.reals is not None:
            oprot.writeFieldBegin('reals', TType.LIST, 1)
            oprot.writeListBegin(TType.LIST, len(self.reals))
            for iter38 in self.reals:
                oprot.writeListBegin(TType.DOUBLE, len(iter38))
                for iter39 in iter38:
                    oprot.writeDouble(iter39)
                oprot.writeListEnd()
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.imags is not None:
            oprot.writeFieldBegin('imags', TType.LIST, 2)
            oprot.writeListBegin(TType.LIST, len(self.imags))
            for iter40 in self.imags:
                oprot.writeListBegin(TType.DOUBLE, len(iter40))
                for iter41 in iter40:
                    oprot.writeDouble(iter41)
                oprot.writeListEnd()
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.unitary is not None:
            oprot.writeFieldBegin('unitary', TType.BOOL, 3)
            oprot.writeBool(self.unitary)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.reals is None:
            raise TProtocolException(message='Required field reals is unset!')
        if self.imags is None:
            raise TProtocolException(message='Required field imags is unset!')
        if self.unitary is None:
            raise TProtocolException(message='Required field unitary is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Cmdex(object):
    """
    Attributes:
     - amp
     - mat

    """


    def __init__(self, amp=None, mat=None,):
        self.amp = amp
        self.mat = mat

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.amp = Amplitude()
                    self.amp.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.mat = Matrix()
                    self.mat.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Cmdex')
        if self.amp is not None:
            oprot.writeFieldBegin('amp', TType.STRUCT, 1)
            self.amp.write(oprot)
            oprot.writeFieldEnd()
        if self.mat is not None:
            oprot.writeFieldBegin('mat', TType.STRUCT, 2)
            self.mat.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Cmd(object):
    """
    Attributes:
     - gate
     - targets
     - controls
     - rotation
     - desc
     - inverse
     - cmdex

    """


    def __init__(self, gate=None, targets=None, controls=None, rotation=None, desc=None, inverse=None, cmdex=None,):
        self.gate = gate
        self.targets = targets
        self.controls = controls
        self.rotation = rotation
        self.desc = desc
        self.inverse = inverse
        self.cmdex = cmdex

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.gate = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.targets = []
                    (_etype45, _size42) = iprot.readListBegin()
                    for _i46 in range(_size42):
                        _elem47 = iprot.readI32()
                        self.targets.append(_elem47)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.controls = []
                    (_etype51, _size48) = iprot.readListBegin()
                    for _i52 in range(_size48):
                        _elem53 = iprot.readI32()
                        self.controls.append(_elem53)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.rotation = []
                    (_etype57, _size54) = iprot.readListBegin()
                    for _i58 in range(_size54):
                        _elem59 = iprot.readDouble()
                        self.rotation.append(_elem59)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.desc = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.BOOL:
                    self.inverse = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.STRUCT:
                    self.cmdex = Cmdex()
                    self.cmdex.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Cmd')
        if self.gate is not None:
            oprot.writeFieldBegin('gate', TType.STRING, 1)
            oprot.writeString(self.gate.encode('utf-8') if sys.version_info[0] == 2 else self.gate)
            oprot.writeFieldEnd()
        if self.targets is not None:
            oprot.writeFieldBegin('targets', TType.LIST, 2)
            oprot.writeListBegin(TType.I32, len(self.targets))
            for iter60 in self.targets:
                oprot.writeI32(iter60)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.controls is not None:
            oprot.writeFieldBegin('controls', TType.LIST, 3)
            oprot.writeListBegin(TType.I32, len(self.controls))
            for iter61 in self.controls:
                oprot.writeI32(iter61)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.rotation is not None:
            oprot.writeFieldBegin('rotation', TType.LIST, 4)
            oprot.writeListBegin(TType.DOUBLE, len(self.rotation))
            for iter62 in self.rotation:
                oprot.writeDouble(iter62)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.desc is not None:
            oprot.writeFieldBegin('desc', TType.STRING, 5)
            oprot.writeString(self.desc.encode('utf-8') if sys.version_info[0] == 2 else self.desc)
            oprot.writeFieldEnd()
        if self.inverse is not None:
            oprot.writeFieldBegin('inverse', TType.BOOL, 6)
            oprot.writeBool(self.inverse)
            oprot.writeFieldEnd()
        if self.cmdex is not None:
            oprot.writeFieldBegin('cmdex', TType.STRUCT, 7)
            self.cmdex.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.gate is None:
            raise TProtocolException(message='Required field gate is unset!')
        if self.targets is None:
            raise TProtocolException(message='Required field targets is unset!')
        if self.controls is None:
            raise TProtocolException(message='Required field controls is unset!')
        if self.rotation is None:
            raise TProtocolException(message='Required field rotation is unset!')
        if self.desc is None:
            raise TProtocolException(message='Required field desc is unset!')
        if self.inverse is None:
            raise TProtocolException(message='Required field inverse is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Circuit(object):
    """
    Attributes:
     - cmds

    """


    def __init__(self, cmds=None,):
        self.cmds = cmds

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.cmds = []
                    (_etype66, _size63) = iprot.readListBegin()
                    for _i67 in range(_size63):
                        _elem68 = Cmd()
                        _elem68.read(iprot)
                        self.cmds.append(_elem68)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Circuit')
        if self.cmds is not None:
            oprot.writeFieldBegin('cmds', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.cmds))
            for iter69 in self.cmds:
                iter69.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.cmds is None:
            raise TProtocolException(message='Required field cmds is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class MeasureResult(object):
    """
    Attributes:
     - id
     - value

    """


    def __init__(self, id=None, value=None,):
        self.id = id
        self.value = value

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.id = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.value = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('MeasureResult')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.I32, 1)
            oprot.writeI32(self.id)
            oprot.writeFieldEnd()
        if self.value is not None:
            oprot.writeFieldBegin('value', TType.I32, 2)
            oprot.writeI32(self.value)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.value is None:
            raise TProtocolException(message='Required field value is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Outcome(object):
    """
    Attributes:
     - bitstr
     - count

    """


    def __init__(self, bitstr=None, count=None,):
        self.bitstr = bitstr
        self.count = count

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.bitstr = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.count = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Outcome')
        if self.bitstr is not None:
            oprot.writeFieldBegin('bitstr', TType.STRING, 1)
            oprot.writeString(self.bitstr.encode('utf-8') if sys.version_info[0] == 2 else self.bitstr)
            oprot.writeFieldEnd()
        if self.count is not None:
            oprot.writeFieldBegin('count', TType.I32, 2)
            oprot.writeI32(self.count)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.bitstr is None:
            raise TProtocolException(message='Required field bitstr is unset!')
        if self.count is None:
            raise TProtocolException(message='Required field count is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Result(object):
    """
    Attributes:
     - measureSet
     - outcomeSet

    """


    def __init__(self, measureSet=None, outcomeSet=None,):
        self.measureSet = measureSet
        self.outcomeSet = outcomeSet

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.measureSet = []
                    (_etype73, _size70) = iprot.readListBegin()
                    for _i74 in range(_size70):
                        _elem75 = MeasureResult()
                        _elem75.read(iprot)
                        self.measureSet.append(_elem75)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.outcomeSet = []
                    (_etype79, _size76) = iprot.readListBegin()
                    for _i80 in range(_size76):
                        _elem81 = Outcome()
                        _elem81.read(iprot)
                        self.outcomeSet.append(_elem81)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Result')
        if self.measureSet is not None:
            oprot.writeFieldBegin('measureSet', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.measureSet))
            for iter82 in self.measureSet:
                iter82.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.outcomeSet is not None:
            oprot.writeFieldBegin('outcomeSet', TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.outcomeSet))
            for iter83 in self.outcomeSet:
                iter83.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.measureSet is None:
            raise TProtocolException(message='Required field measureSet is unset!')
        if self.outcomeSet is None:
            raise TProtocolException(message='Required field outcomeSet is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class InitQubitsReq(object):
    """
    Attributes:
     - id
     - qubits
     - density
     - exec_type
     - hosts

    """


    def __init__(self, id=None, qubits=None, density=None, exec_type=None, hosts=None,):
        self.id = id
        self.qubits = qubits
        self.density = density
        self.exec_type = exec_type
        self.hosts = hosts

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.qubits = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.BOOL:
                    self.density = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I32:
                    self.exec_type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.LIST:
                    self.hosts = []
                    (_etype87, _size84) = iprot.readListBegin()
                    for _i88 in range(_size84):
                        _elem89 = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                        self.hosts.append(_elem89)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('InitQubitsReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.qubits is not None:
            oprot.writeFieldBegin('qubits', TType.I32, 2)
            oprot.writeI32(self.qubits)
            oprot.writeFieldEnd()
        if self.density is not None:
            oprot.writeFieldBegin('density', TType.BOOL, 3)
            oprot.writeBool(self.density)
            oprot.writeFieldEnd()
        if self.exec_type is not None:
            oprot.writeFieldBegin('exec_type', TType.I32, 4)
            oprot.writeI32(self.exec_type)
            oprot.writeFieldEnd()
        if self.hosts is not None:
            oprot.writeFieldBegin('hosts', TType.LIST, 5)
            oprot.writeListBegin(TType.STRING, len(self.hosts))
            for iter90 in self.hosts:
                oprot.writeString(iter90.encode('utf-8') if sys.version_info[0] == 2 else iter90)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.qubits is None:
            raise TProtocolException(message='Required field qubits is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class InitQubitsResp(object):
    """
    Attributes:
     - base

    """


    def __init__(self, base=None,):
        self.base = base

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('InitQubitsResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SendCircuitCmdReq(object):
    """
    Attributes:
     - id
     - circuit
     - final

    """


    def __init__(self, id=None, circuit=None, final=None,):
        self.id = id
        self.circuit = circuit
        self.final = final

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.circuit = Circuit()
                    self.circuit.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.BOOL:
                    self.final = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('SendCircuitCmdReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.circuit is not None:
            oprot.writeFieldBegin('circuit', TType.STRUCT, 2)
            self.circuit.write(oprot)
            oprot.writeFieldEnd()
        if self.final is not None:
            oprot.writeFieldBegin('final', TType.BOOL, 3)
            oprot.writeBool(self.final)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SendCircuitCmdResp(object):
    """
    Attributes:
     - base

    """


    def __init__(self, base=None,):
        self.base = base

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('SendCircuitCmdResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class CancelCmdReq(object):
    """
    Attributes:
     - id

    """


    def __init__(self, id=None,):
        self.id = id

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('CancelCmdReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class CancelCmdResp(object):
    """
    Attributes:
     - base

    """


    def __init__(self, base=None,):
        self.base = base

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('CancelCmdResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetProbAmpReq(object):
    """
    Attributes:
     - id
     - index

    """


    def __init__(self, id=None, index=None,):
        self.id = id
        self.index = index

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I64:
                    self.index = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetProbAmpReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.index is not None:
            oprot.writeFieldBegin('index', TType.I64, 2)
            oprot.writeI64(self.index)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.index is None:
            raise TProtocolException(message='Required field index is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetProbAmpResp(object):
    """
    Attributes:
     - base
     - amp

    """


    def __init__(self, base=None, amp=None,):
        self.base = base
        self.amp = amp

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.DOUBLE:
                    self.amp = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetProbAmpResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.amp is not None:
            oprot.writeFieldBegin('amp', TType.DOUBLE, 2)
            oprot.writeDouble(self.amp)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetProbOfOutcomeReq(object):
    """
    Attributes:
     - id
     - qubit
     - outcom

    """


    def __init__(self, id=None, qubit=None, outcom=None,):
        self.id = id
        self.qubit = qubit
        self.outcom = outcom

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.qubit = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.outcom = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetProbOfOutcomeReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.qubit is not None:
            oprot.writeFieldBegin('qubit', TType.I32, 2)
            oprot.writeI32(self.qubit)
            oprot.writeFieldEnd()
        if self.outcom is not None:
            oprot.writeFieldBegin('outcom', TType.I32, 3)
            oprot.writeI32(self.outcom)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.qubit is None:
            raise TProtocolException(message='Required field qubit is unset!')
        if self.outcom is None:
            raise TProtocolException(message='Required field outcom is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetProbOfOutcomeResp(object):
    """
    Attributes:
     - base
     - pro_outcome

    """


    def __init__(self, base=None, pro_outcome=None,):
        self.base = base
        self.pro_outcome = pro_outcome

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.DOUBLE:
                    self.pro_outcome = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetProbOfOutcomeResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.pro_outcome is not None:
            oprot.writeFieldBegin('pro_outcome', TType.DOUBLE, 2)
            oprot.writeDouble(self.pro_outcome)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetProbOfAllOutcomReq(object):
    """
    Attributes:
     - id
     - targets

    """


    def __init__(self, id=None, targets=None,):
        self.id = id
        self.targets = targets

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.targets = []
                    (_etype94, _size91) = iprot.readListBegin()
                    for _i95 in range(_size91):
                        _elem96 = iprot.readI32()
                        self.targets.append(_elem96)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetProbOfAllOutcomReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.targets is not None:
            oprot.writeFieldBegin('targets', TType.LIST, 2)
            oprot.writeListBegin(TType.I32, len(self.targets))
            for iter97 in self.targets:
                oprot.writeI32(iter97)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.targets is None:
            raise TProtocolException(message='Required field targets is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetProbOfAllOutcomResp(object):
    """
    Attributes:
     - base
     - pro_outcomes

    """


    def __init__(self, base=None, pro_outcomes=None,):
        self.base = base
        self.pro_outcomes = pro_outcomes

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.pro_outcomes = []
                    (_etype101, _size98) = iprot.readListBegin()
                    for _i102 in range(_size98):
                        _elem103 = iprot.readDouble()
                        self.pro_outcomes.append(_elem103)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetProbOfAllOutcomResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.pro_outcomes is not None:
            oprot.writeFieldBegin('pro_outcomes', TType.LIST, 2)
            oprot.writeListBegin(TType.DOUBLE, len(self.pro_outcomes))
            for iter104 in self.pro_outcomes:
                oprot.writeDouble(iter104)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetAllStateReq(object):
    """
    Attributes:
     - id

    """


    def __init__(self, id=None,):
        self.id = id

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetAllStateReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetAllStateResp(object):
    """
    Attributes:
     - base
     - all_state

    """


    def __init__(self, base=None, all_state=None,):
        self.base = base
        self.all_state = all_state

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.all_state = []
                    (_etype108, _size105) = iprot.readListBegin()
                    for _i109 in range(_size105):
                        _elem110 = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                        self.all_state.append(_elem110)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetAllStateResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.all_state is not None:
            oprot.writeFieldBegin('all_state', TType.LIST, 2)
            oprot.writeListBegin(TType.STRING, len(self.all_state))
            for iter111 in self.all_state:
                oprot.writeString(iter111.encode('utf-8') if sys.version_info[0] == 2 else iter111)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ApplyQFTReq(object):
    """
    Attributes:
     - id
     - targets

    """


    def __init__(self, id=None, targets=None,):
        self.id = id
        self.targets = targets

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.targets = []
                    (_etype115, _size112) = iprot.readListBegin()
                    for _i116 in range(_size112):
                        _elem117 = iprot.readI32()
                        self.targets.append(_elem117)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ApplyQFTReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.targets is not None:
            oprot.writeFieldBegin('targets', TType.LIST, 2)
            oprot.writeListBegin(TType.I32, len(self.targets))
            for iter118 in self.targets:
                oprot.writeI32(iter118)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.targets is None:
            raise TProtocolException(message='Required field targets is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ApplyQFTResp(object):
    """
    Attributes:
     - base

    """


    def __init__(self, base=None,):
        self.base = base

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ApplyQFTResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ApplyFullQFTReq(object):
    """
    Attributes:
     - id

    """


    def __init__(self, id=None,):
        self.id = id

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ApplyFullQFTReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ApplyFullQFTResp(object):
    """
    Attributes:
     - base

    """


    def __init__(self, base=None,):
        self.base = base

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ApplyFullQFTResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class RunCircuitReq(object):
    """
    Attributes:
     - id
     - shots

    """


    def __init__(self, id=None, shots=None,):
        self.id = id
        self.shots = shots

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.shots = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('RunCircuitReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.shots is not None:
            oprot.writeFieldBegin('shots', TType.I32, 2)
            oprot.writeI32(self.shots)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.shots is None:
            raise TProtocolException(message='Required field shots is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class RunCircuitResp(object):
    """
    Attributes:
     - base
     - result

    """


    def __init__(self, base=None, result=None,):
        self.base = base
        self.result = result

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.result = Result()
                    self.result.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('RunCircuitResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.result is not None:
            oprot.writeFieldBegin('result', TType.STRUCT, 2)
            self.result.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetExpecPauliProdReq(object):
    """
    Attributes:
     - id
     - pauli_prod

    """


    def __init__(self, id=None, pauli_prod=None,):
        self.id = id
        self.pauli_prod = pauli_prod

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.pauli_prod = []
                    (_etype122, _size119) = iprot.readListBegin()
                    for _i123 in range(_size119):
                        _elem124 = PauliProdInfo()
                        _elem124.read(iprot)
                        self.pauli_prod.append(_elem124)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetExpecPauliProdReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.pauli_prod is not None:
            oprot.writeFieldBegin('pauli_prod', TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.pauli_prod))
            for iter125 in self.pauli_prod:
                iter125.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.pauli_prod is None:
            raise TProtocolException(message='Required field pauli_prod is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetExpecPauliProdResp(object):
    """
    Attributes:
     - base
     - expect

    """


    def __init__(self, base=None, expect=None,):
        self.base = base
        self.expect = expect

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.DOUBLE:
                    self.expect = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetExpecPauliProdResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.expect is not None:
            oprot.writeFieldBegin('expect', TType.DOUBLE, 2)
            oprot.writeDouble(self.expect)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class PauliProdInfo(object):
    """
    Attributes:
     - oper_type
     - target

    """


    def __init__(self, oper_type=None, target=None,):
        self.oper_type = oper_type
        self.target = target

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.oper_type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.target = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('PauliProdInfo')
        if self.oper_type is not None:
            oprot.writeFieldBegin('oper_type', TType.I32, 1)
            oprot.writeI32(self.oper_type)
            oprot.writeFieldEnd()
        if self.target is not None:
            oprot.writeFieldBegin('target', TType.I32, 2)
            oprot.writeI32(self.target)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.oper_type is None:
            raise TProtocolException(message='Required field oper_type is unset!')
        if self.target is None:
            raise TProtocolException(message='Required field target is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetExpecPauliSumReq(object):
    """
    Attributes:
     - id
     - oper_type_list
     - term_coeff_list

    """


    def __init__(self, id=None, oper_type_list=None, term_coeff_list=None,):
        self.id = id
        self.oper_type_list = oper_type_list
        self.term_coeff_list = term_coeff_list

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.id = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.oper_type_list = []
                    (_etype129, _size126) = iprot.readListBegin()
                    for _i130 in range(_size126):
                        _elem131 = iprot.readI32()
                        self.oper_type_list.append(_elem131)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.term_coeff_list = []
                    (_etype135, _size132) = iprot.readListBegin()
                    for _i136 in range(_size132):
                        _elem137 = iprot.readDouble()
                        self.term_coeff_list.append(_elem137)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetExpecPauliSumReq')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.STRING, 1)
            oprot.writeString(self.id.encode('utf-8') if sys.version_info[0] == 2 else self.id)
            oprot.writeFieldEnd()
        if self.oper_type_list is not None:
            oprot.writeFieldBegin('oper_type_list', TType.LIST, 2)
            oprot.writeListBegin(TType.I32, len(self.oper_type_list))
            for iter138 in self.oper_type_list:
                oprot.writeI32(iter138)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.term_coeff_list is not None:
            oprot.writeFieldBegin('term_coeff_list', TType.LIST, 3)
            oprot.writeListBegin(TType.DOUBLE, len(self.term_coeff_list))
            for iter139 in self.term_coeff_list:
                oprot.writeDouble(iter139)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.id is None:
            raise TProtocolException(message='Required field id is unset!')
        if self.oper_type_list is None:
            raise TProtocolException(message='Required field oper_type_list is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GetExpecPauliSumResp(object):
    """
    Attributes:
     - base
     - expect

    """


    def __init__(self, base=None, expect=None,):
        self.base = base
        self.expect = expect

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.base = qutrunk.sim.qusprout.code.ttypes.BaseCode()
                    self.base.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.DOUBLE:
                    self.expect = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('GetExpecPauliSumResp')
        if self.base is not None:
            oprot.writeFieldBegin('base', TType.STRUCT, 1)
            self.base.write(oprot)
            oprot.writeFieldEnd()
        if self.expect is not None:
            oprot.writeFieldBegin('expect', TType.DOUBLE, 2)
            oprot.writeDouble(self.expect)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.base is None:
            raise TProtocolException(message='Required field base is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(Amplitude)
Amplitude.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'reals', (TType.DOUBLE, None, False), None, ),  # 1
    (2, TType.LIST, 'imags', (TType.DOUBLE, None, False), None, ),  # 2
    (3, TType.I32, 'startind', None, None, ),  # 3
    (4, TType.I32, 'numamps', None, None, ),  # 4
)
all_structs.append(Matrix)
Matrix.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'reals', (TType.LIST, (TType.DOUBLE, None, False), False), None, ),  # 1
    (2, TType.LIST, 'imags', (TType.LIST, (TType.DOUBLE, None, False), False), None, ),  # 2
    (3, TType.BOOL, 'unitary', None, None, ),  # 3
)
all_structs.append(Cmdex)
Cmdex.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'amp', [Amplitude, None], None, ),  # 1
    (2, TType.STRUCT, 'mat', [Matrix, None], None, ),  # 2
)
all_structs.append(Cmd)
Cmd.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'gate', 'UTF8', None, ),  # 1
    (2, TType.LIST, 'targets', (TType.I32, None, False), None, ),  # 2
    (3, TType.LIST, 'controls', (TType.I32, None, False), None, ),  # 3
    (4, TType.LIST, 'rotation', (TType.DOUBLE, None, False), None, ),  # 4
    (5, TType.STRING, 'desc', 'UTF8', None, ),  # 5
    (6, TType.BOOL, 'inverse', None, None, ),  # 6
    (7, TType.STRUCT, 'cmdex', [Cmdex, None], None, ),  # 7
)
all_structs.append(Circuit)
Circuit.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'cmds', (TType.STRUCT, [Cmd, None], False), None, ),  # 1
)
all_structs.append(MeasureResult)
MeasureResult.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'id', None, None, ),  # 1
    (2, TType.I32, 'value', None, None, ),  # 2
)
all_structs.append(Outcome)
Outcome.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'bitstr', 'UTF8', None, ),  # 1
    (2, TType.I32, 'count', None, None, ),  # 2
)
all_structs.append(Result)
Result.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'measureSet', (TType.STRUCT, [MeasureResult, None], False), None, ),  # 1
    (2, TType.LIST, 'outcomeSet', (TType.STRUCT, [Outcome, None], False), None, ),  # 2
)
all_structs.append(InitQubitsReq)
InitQubitsReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.I32, 'qubits', None, None, ),  # 2
    (3, TType.BOOL, 'density', None, None, ),  # 3
    (4, TType.I32, 'exec_type', None, None, ),  # 4
    (5, TType.LIST, 'hosts', (TType.STRING, 'UTF8', False), None, ),  # 5
)
all_structs.append(InitQubitsResp)
InitQubitsResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
)
all_structs.append(SendCircuitCmdReq)
SendCircuitCmdReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'circuit', [Circuit, None], None, ),  # 2
    (3, TType.BOOL, 'final', None, None, ),  # 3
)
all_structs.append(SendCircuitCmdResp)
SendCircuitCmdResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
)
all_structs.append(CancelCmdReq)
CancelCmdReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
)
all_structs.append(CancelCmdResp)
CancelCmdResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
)
all_structs.append(GetProbAmpReq)
GetProbAmpReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.I64, 'index', None, None, ),  # 2
)
all_structs.append(GetProbAmpResp)
GetProbAmpResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.DOUBLE, 'amp', None, None, ),  # 2
)
all_structs.append(GetProbOfOutcomeReq)
GetProbOfOutcomeReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.I32, 'qubit', None, None, ),  # 2
    (3, TType.I32, 'outcom', None, None, ),  # 3
)
all_structs.append(GetProbOfOutcomeResp)
GetProbOfOutcomeResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.DOUBLE, 'pro_outcome', None, None, ),  # 2
)
all_structs.append(GetProbOfAllOutcomReq)
GetProbOfAllOutcomReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.LIST, 'targets', (TType.I32, None, False), None, ),  # 2
)
all_structs.append(GetProbOfAllOutcomResp)
GetProbOfAllOutcomResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.LIST, 'pro_outcomes', (TType.DOUBLE, None, False), None, ),  # 2
)
all_structs.append(GetAllStateReq)
GetAllStateReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
)
all_structs.append(GetAllStateResp)
GetAllStateResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.LIST, 'all_state', (TType.STRING, 'UTF8', False), None, ),  # 2
)
all_structs.append(ApplyQFTReq)
ApplyQFTReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.LIST, 'targets', (TType.I32, None, False), None, ),  # 2
)
all_structs.append(ApplyQFTResp)
ApplyQFTResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
)
all_structs.append(ApplyFullQFTReq)
ApplyFullQFTReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
)
all_structs.append(ApplyFullQFTResp)
ApplyFullQFTResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
)
all_structs.append(RunCircuitReq)
RunCircuitReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.I32, 'shots', None, None, ),  # 2
)
all_structs.append(RunCircuitResp)
RunCircuitResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.STRUCT, 'result', [Result, None], None, ),  # 2
)
all_structs.append(GetExpecPauliProdReq)
GetExpecPauliProdReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.LIST, 'pauli_prod', (TType.STRUCT, [PauliProdInfo, None], False), None, ),  # 2
)
all_structs.append(GetExpecPauliProdResp)
GetExpecPauliProdResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.DOUBLE, 'expect', None, None, ),  # 2
)
all_structs.append(PauliProdInfo)
PauliProdInfo.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'oper_type', None, None, ),  # 1
    (2, TType.I32, 'target', None, None, ),  # 2
)
all_structs.append(GetExpecPauliSumReq)
GetExpecPauliSumReq.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'id', 'UTF8', None, ),  # 1
    (2, TType.LIST, 'oper_type_list', (TType.I32, None, False), None, ),  # 2
    (3, TType.LIST, 'term_coeff_list', (TType.DOUBLE, None, False), None, ),  # 3
)
all_structs.append(GetExpecPauliSumResp)
GetExpecPauliSumResp.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'base', [qutrunk.sim.qusprout.code.ttypes.BaseCode, None], None, ),  # 1
    (2, TType.DOUBLE, 'expect', None, None, ),  # 2
)
fix_spec(all_structs)
del all_structs



import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class SqrtXGate(BasicGate):
    """Square-root X gate class.

    Example:
        .. code-block:: python

            SqrtX * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "SqrtX"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply SqrtX gate.

        Example:
            .. code-block:: python

                SqrtX * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return 0.5 * np.array([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]])

    def inv(self):
        """Apply inverse gate."""
        gate = SqrtXGate()
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("SqrtX gate do not support multiple control bits.")
        gate = CSqrtXGate()
        gate.is_inverse = self.is_inverse
        return gate


SqrtX = SqrtXGate()


class CSqrtXGate(BasicGate):
    """Controlled-√X gate.

    Example:
        .. code-block:: python

            CSqrtX * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "CSqrtX"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            .. code-block:: python

                CSqrtX * (qr[0], qr[1])

        Raises:
            TypeError: If the argument is not a Qubit object.
            AttributeError: If the qubits should not be two.
        """
        if not all(isinstance(qubit, QuBit) for qubit in qubits):
            raise TypeError("The argument must be Qubit object.")

        if len(qubits) != 2:
            raise ValueError(
                "Parameter error: One controlled and one target qubit are required."
            )

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.array(
            [(1 + 1j) / 2, 0, (1 - 1j) / 2, 0],
            [0, 1, 0, 0],
            [(1 - 1j) / 2, 0, (1 + 1j) / 2, 0],
            [0, 0, 0, 1],
        )

    def inv(self):
        """Apply inverse gate"""
        gate = CSqrtXGate()
        gate.is_inverse = not self.is_inverse
        return gate


CSqrtX = CSqrtXGate()



from enum import Enum


class BasicGate:
    """Base class of all gates. (Don't use it directly but derive from it)."""

    def __init__(self):
        self.is_inverse = False
        self.name = ""

    # TODO: need to improve.
    def __str__(self):
        """The string description of a quantum logic gate."""
        return ""

    # TODO: need to improve.
    def __or__(self, qubit):
        """Quantum logic gate operation. Overwrite | operator to achieve quantum logic gate operation."""
        pass

    # TODO: need to improve.
    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        pass

    def angles(self):
        """Get angle Args."""
        angle_params = ("rotation", "theta", "phi", "lam", "gamma")
        return [getattr(self, param) for param in angle_params if hasattr(self, param)]

    def commit(self, circuit, cmd):
        """Commit command to circuit.

        Args:
            circuit: The circuit of cmd to committed.
            cmd: The target command committed to circuit.
            statement: The origin statement that generate target cmd.
        """
        circuit.append_cmd(cmd)

    # TODO: need to improve.
    def inv(self):
        """Apply inverse gate."""
        # note: 状态相关方法需要重新生成一个新对象
        raise NotImplementedError

    # TODO: need to improve.
    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        # note: 状态相关方法需要重新生成一个新对象
        raise NotImplementedError


class BasicRotateGate(BasicGate):
    """Base class for rotation gate."""

    def __init__(self):
        super().__init__()


class BasicPhaseGate(BasicGate):
    """Base class for phase gate."""

    def __init__(self):
        super().__init__()


# TODO: need to improve.
class Observable:
    """Base class representing observables. It's usually PauliX, PauliY, PauliZ, PauliI."""

    def __call__(self, target):
        """
        Get Observable data.

        Args:
            target: The observed qubit.

        Returns:
            The observed data list, each item contains op type and target qubit, \
                e.g: [{"oper_type": 1, "target": 0}]
        """
        return []


class PauliType(Enum):
    """PauliType."""

    PAULI_I = 0
    PAULI_X = 1
    PAULI_Y = 2
    PAULI_Z = 3

# todo: 用于封装circuit.expval_pauli方法的参数调用
class PauliOps:
    """A list of Pauli Operator."""
    def __init__(self):
        self.pauli_ops = []

    def append(self, op):
        """Append Pauli Operator."""
        self.pauli_ops.append(op)

    def __lshift__(self, op):
        self.append(op)
        return self

    def __getitem__(self, index):
        return self.pauli_ops[index]

class PauliCoeff:
    """Pauli Coeff consist of Pauli operator and coefficients."""
    def __init__(self, coeff, paulis):
        self.coeff = coeff
        self.paulis = [op.value for op in paulis]

    def padding(self, len, op=PauliType.PAULI_I):
        """Fill paulis by specific pauli operator in len times.
        
        Args:
            len: Times to fill op.
            op: The target operator filled to paulis, defaul: PauliI.
        """
        for i in range(len):
            self.paulis.append(op.value)

class PauliCoeffs:
    """A list of PauliCoeff."""
    def __init__(self):
        self.pauli_term_list = []

    def append(self, term):
        """Append PaulCoeff term."""
        self.pauli_term_list.append(term)

    def __lshift__(self, term):
        self.append(term)
        return self

    def __getitem__(self, index):
        return self.pauli_term_list[index]
        



"""Exceptions for errors raised by local backend."""

from qutrunk.exceptions import QuTrunkError


class LocalBackendError(QuTrunkError):
    """Base class for errors raised by Local backend."""

    def __init__(self, *message):
        """Set the error message."""
        super().__init__(*message)
        self.message = " ".join(message)

    def __str__(self):
        """Return the message."""
        return repr(self.message)



"""All gate operations supported by qutrunk."""

from math import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import *


@Gate
def my_gate(a, b):
    return Gate() << (H, a) << (CNOT, (a, b))


def run_gates():
    # Create and allocate quantum circuit
    qc = QCircuit()
    qr = qc.allocate(3)

    # Apply quantum gate
    H * qr[0]
    CNOT * (qr[0], qr[1])
    NOT * qr[0]
    Toffoli * (qr[0], qr[1], qr[2])
    P(pi / 2) * qr[2]
    R(pi / 2, pi / 2) * qr[0]
    Rx(pi / 2) * qr[1]
    Ry(pi / 2) * qr[1]
    Rz(pi / 2) * qr[1]
    S * qr[0]
    Sdg * qr[0]
    T * qr[0]
    Tdg * qr[0]
    X * qr[2]
    Y * qr[2]
    Z * qr[2]
    X1 * qr[0]
    Y1 * qr[0]
    Z1 * qr[0]
    Swap * (qr[0], qr[1])
    iSwap * (qr[0], qr[1])
    SqrtX * qr[0]

    CX * (qr[0], qr[1])
    CY * (qr[0], qr[1])
    CZ * (qr[0], qr[1])
    CP(pi / 2) * (qr[0], qr[1])
    CR(pi / 2) * (qr[0], qr[1])
    CRx(pi / 2) * (qr[0], qr[1])
    CRy(pi / 2) * (qr[0], qr[1])
    CRz(pi / 2) * (qr[0], qr[1])
    MCX(2) * (qr[0], qr[1], qr[2])
    MCZ(2) * (qr[0], qr[1], qr[2])

    Rxx(pi / 2) * (qr[0], qr[1])
    Ryy(pi / 2) * (qr[0], qr[1])
    Rzz(pi / 2) * (qr[0], qr[1])

    U1(pi / 2) * qr[0]
    U2(pi / 2, pi / 2) * qr[0]
    U3(pi / 2, pi / 2, pi / 2) * qr[0]
    CU(pi / 2, pi / 2, pi / 2, pi / 2) * (qr[0], qr[1])
    CU1(pi / 2) * (qr[1], qr[2])
    CU3(pi / 2, pi / 2, pi / 2) * (qr[0], qr[1])
    I * qr[0]

    CH * (qr[0], qr[1])
    CSwap * (qr[0], qr[1], qr[2])
    CSqrtX * (qr[0], qr[1])
    SqrtXdg * qr[0]

    Barrier * qr

    Power(2, H) * qr[0]

    my_gate * (qr[0], qr[1])
    Barrier * qr
    Power(2, my_gate) * (qr[0], qr[1])

    # Measure all quantum qubits
    All(Measure) * qr

    # Print quantum circuit
    qc.print()

    # Print quantum circuit as operqasm grammar
    qc.print(format="openqasm")

    # Run quantum circuit
    qc.run()

    return qc


if __name__ == "__main__":
    # Run locally
    circuit = run_gates()




import math

from qutrunk.test.gate.global_parameters import PRECISION


def check_all_state_inverse(res, res_box):
    if len(res) != len(res_box):
        return False

    for index in range(len(res)):
        amp_str = res[index]
        amp_str_box = res_box[index]
        real_str, image_str = amp_str.split(",")
        real_str_box, image_str_box = amp_str_box.split(",")
        test = float(image_str) - float(image_str_box)
        test1 = math.fabs(test)
        a = test1 > PRECISION
        if (
            math.fabs(float(real_str) - float(real_str_box)) > PRECISION
            or math.fabs(float(image_str) - float(image_str_box)) > PRECISION
        ):
            return False

    return True



import os
import argparse


def set_config(yaml_file, json_str):
    """Write the json to file config/qubox.yaml.

    Args:
        yaml_file: Yaml file path.
        json_str: Json content.
    """
    # 将json转成要写入的内容
    file_str = ""
    for key in json_str:
        value = json_str[key]
        file_str += f"{key}: {value}\n"

    # 修改yaml文件内容
    with open(yaml_file, "w", encoding="utf-8") as f:
        f.write(file_str)


def str_to_dict(args_json):
    """Convert strng to dict.

    Args:
        args_json: Json string.

    Return:
        Dict object.
    """
    json_str = {}
    args_json = args_json.split(":")
    json_str["ip"], json_str["port"] = args_json[0], int(args_json[1])
    return json_str


def set_box_ip(args_json=None):
    """Set Qubox IP.

    Args:
        args_json: QuBox config.

    Raises:
        Exception: If no parameters are passed in.
    """
    # 1 调用该函数，接收参数。
    if args_json:  # 判断是否传入参数
        # 2 读取传入字符内容并转为JSON格式

        # 3 修改config下面的yaml文件
        BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        yaml_file = os.path.join(BASE_DIR, "config", "qubox.yaml")  # 获取yaml文件路径
        json_str = str_to_dict(args_json)  # 获取json
        set_config(yaml_file, json_str)
        print("Done")
    else:
        raise Exception("未传参数")


if __name__ == "__main__":
    # 接收参数
    parser = argparse.ArgumentParser()
    parser.add_argument("--ip", type=str)
    args = parser.parse_args()
    # 执行修改
    set_box_ip((args.ip).replace(" ", "").replace("：", ":"))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Exception for errors raised while interpreting nodes."""


class NodeException(Exception):
    """Base class for errors raised while interpreting nodes."""

    def __init__(self, *msg):
        """Set the error message."""
        super().__init__(*msg)
        self.msg = " ".join(msg)

    def __str__(self):
        """Return the message."""
        return repr(self.msg)



"""A layer is the column of the circuit."""
from .text_draw_element import (
    BoxOnQuWireTop,
    BoxOnQuWireMid,
    BoxOnClWireMid,
    BoxOnClWireBot,
    BoxOnClWire,
    BoxOnClWireTop,
    BoxOnQuWire,
    BoxOnQuWireBot,
    Bullet,
    OpenBullet,
)
from qutrunk.circuit import CBit, CReg, QuBit


class Layer:
    """A layer is the column of the circuit."""

    def __init__(self, qubits, cbits, circuit=None):
        # QuBit object list
        self.qubits = qubits
        # CBit object list
        self.cbits = cbits
        # quantum circuit
        self.circuit = circuit
        # quantum layer
        self.qubit_layer = [None] * len(qubits)
        # connections
        self.connections = []
        # classics bit layer
        self.cbit_layer = [None] * len(cbits)

    @property
    def full_layer(self):
        """Returns the composition of qubits and classic wires."""
        return self.qubit_layer + self.cbit_layer

    def set_qubit(self, qubit, element):
        """Sets the qubit to the element.

        Args:
            qubit: Element of self.qubits.
            element: Element to draw.
        """
        self.qubit_layer[self.qubits.index(qubit)] = element

    def set_cbit(self, cbit, element):
        """Sets the cbit to the element."""

        self.cbit_layer[0] = element

    def _set_multibox(
        self,
        label,
        qubits=None,
        cbits=None,
        top_connect=None,
        bot_connect=None,
        conditional=False,
        controlled_edge=None,
    ):
        # both of qubits and cbits is not none
        if qubits is not None and cbits is not None:
            # Gets the subscript of the cbit element
            cbit_index = sorted(i for i, x in enumerate(self.cbits) if x in cbits)
            # Gets the subscript of the qubit element
            qbit_index = sorted(i for i, x in enumerate(self.qubits) if x in qubits)

            # Use the element subscript as the label and find the longest subscript length
            wire_label_len = max(len(str(len(qubits) - 1)), len(str(len(cbits) - 1)))

            # Returns a list of left-aligned strings with qubit subscripts
            qargs = [
                str(qubits.index(qbit)).ljust(wire_label_len, " ")
                for qbit in self.qubits
                if qbit in qubits
            ]
            # Returns a list of left-aligned strings with cbit subscripts
            cargs = [
                str(cbits.index(cbit)).ljust(wire_label_len, " ")
                for cbit in self.cbits
                if cbit in cbits
            ]
            # order by index
            qubits = sorted(qubits, key=self.qubits.index)
            cbits = sorted(cbits, key=self.cbits.index)

            # box height
            box_height = len(self.qubits) - min(qbit_index) + max(cbit_index) + 1

            # set top element
            self.set_qubit(
                qubits.pop(0), BoxOnQuWireTop(label, wire_label=qargs.pop(0))
            )
            order = 0
            for order, bit_i in enumerate(range(min(qbit_index) + 1, len(self.qubits))):
                if bit_i in qbit_index:
                    named_bit = qubits.pop(0)
                    wire_label = qargs.pop(0)
                else:
                    named_bit = self.qubits[bit_i]
                    wire_label = " " * wire_label_len
                # set middle element
                self.set_qubit(
                    named_bit,
                    BoxOnQuWireMid(label, box_height, order, wire_label=wire_label),
                )

            # classics bit
            for order, bit_i in enumerate(range(max(cbit_index)), order + 1):
                if bit_i in cbit_index:
                    named_bit = cbits.pop(0)
                    wire_label = cargs.pop(0)
                else:
                    named_bit = self.cbits[bit_i]
                    wire_label = " " * wire_label_len
                self.set_cbit(
                    named_bit,
                    BoxOnClWireMid(label, box_height, order, wire_label=wire_label),
                )
            self.set_cbit(
                cbits.pop(0), BoxOnClWireBot(label, box_height, wire_label=cargs.pop(0))
            )
            return cbit_index

        # qubits is none,but cbits is bot
        if qubits is None and cbits is not None:
            bits = list(cbits)
            bit_index = sorted(i for i, x in enumerate(self.cbits) if x in bits)
            wire_label_len = len(str(len(bits) - 1))
            bits.sort(key=self.cbits.index)
            qargs = [""] * len(bits)
            set_bit = self.set_cbit
            OnWire = BoxOnClWire
            OnWireTop = BoxOnClWireTop
            OnWireMid = BoxOnClWireMid
            OnWireBot = BoxOnClWireBot
        # cbits is none,but qubits is not
        elif cbits is None and qubits is not None:
            bits = list(qubits)
            bit_index = sorted(i for i, x in enumerate(self.qubits) if x in bits)
            wire_label_len = len(str(len(bits) - 1))
            qargs = [
                str(bits.index(qbit)).ljust(wire_label_len, " ")
                for qbit in self.qubits
                if qbit in bits
            ]
            bits.sort(key=self.qubits.index)
            set_bit = self.set_qubit
            OnWire = BoxOnQuWire
            OnWireTop = BoxOnQuWireTop
            OnWireMid = BoxOnQuWireMid
            OnWireBot = BoxOnQuWireBot
        else:
            raise ValueError("_set_multibox error.")

        control_index = {}
        if controlled_edge:
            for index, qubit in enumerate(self.qubits):
                for qubit_in_edge, value in controlled_edge:
                    if qubit == qubit_in_edge:
                        control_index[index] = "■" if value == "1" else "o"

        if len(bit_index) == 1:
            set_bit(bits[0], OnWire(label, top_connect=top_connect))
        else:
            box_height = max(bit_index) - min(bit_index) + 1
            set_bit(
                bits.pop(0),
                OnWireTop(label, top_connect=top_connect, wire_label=qargs.pop(0)),
            )

            for order, bit_i in enumerate(range(min(bit_index) + 1, max(bit_index))):
                if bit_i in bit_index:
                    named_bit = bits.pop(0)
                    wire_label = qargs.pop(0)
                else:
                    named_bit = (self.qubits + self.cbits)[bit_i]
                    wire_label = " " * wire_label_len

                control_label = control_index.get(bit_i)
                set_bit(
                    named_bit,
                    OnWireMid(
                        label,
                        box_height,
                        order,
                        wire_label=wire_label,
                        control_label=control_label,
                    ),
                )

            set_bit(
                bits.pop(0),
                OnWireBot(
                    label,
                    box_height,
                    bot_connect=bot_connect,
                    wire_label=qargs.pop(0),
                    conditional=conditional,
                ),
            )
        return bit_index

    def set_qu_multibox(
        self,
        bits,
        label,
        top_connect=None,
        bot_connect=None,
        conditional=False,
        controlled_edge=None,
    ):
        """Sets the multi qubit box."""
        return self._set_multibox(
            label,
            qubits=bits,
            top_connect=top_connect,
            bot_connect=bot_connect,
            conditional=conditional,
            controlled_edge=controlled_edge,
        )

    def connect_with(self, wire_char):
        """Connects the elements in the layer using wire_char."""
        if len([qbit for qbit in self.qubit_layer if qbit is not None]) == 1:
            return None

        for label, affected_bits in self.connections:

            if not affected_bits:
                continue

            affected_bits[0].connect(wire_char, ["bot"])
            for affected_bit in affected_bits[1:-1]:
                affected_bit.connect(wire_char, ["bot", "top"])

            affected_bits[-1].connect(wire_char, ["top"], label)

            if label:
                for affected_bit in affected_bits:
                    affected_bit.right_fill = len(label) + len(affected_bit.mid)
                    if (
                        isinstance(affected_bit, (Bullet, OpenBullet))
                        and affected_bit.conditional
                    ):
                        affected_bit.left_fill = len(label) + len(affected_bit.mid)

    def set_cl_multibox(self, condition, top_connect="┴"):
        """set multi cbit box."""
        pass


def get_condition_label_val(condition, circuit, cregbundle=True, reverse_bits=False):
    """Get the label and value list to display a condition."""
    cond_is_bit = bool(isinstance(condition[0], CBit))
    cond_val = int(condition[1])

    if isinstance(condition[0], CReg) and not cregbundle:
        val_bits = list(f"{cond_val:0{condition[0].size}b}")
        if not reverse_bits:
            val_bits = val_bits[::-1]
    else:
        val_bits = list(str(cond_val))

    label = ""
    if cond_is_bit and cregbundle:
        register, _, reg_index = get_bit_reg_index(circuit, condition[0])
        if register is not None:
            label = f"{register.name}_{reg_index}={hex(cond_val)}"
    elif not cond_is_bit:
        label = hex(cond_val)

    return label, val_bits


def get_bit_reg_index(circuit, bit):
    """Gets the bit subscript and the corresponding register."""

    # get subscript of bit
    bit_index = circuit.find_bit(bit)
    reg_index = bit_index
    # Get the corresponding register
    if isinstance(bit, QuBit):
        register = circuit.qreg[bit]

    if isinstance(bit, CBit):
        register = circuit.creg[bit]

    return register, bit_index, reg_index



import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class SqrtXdgGate(BasicGate):
    """The inverse single-qubit Sqrt(X) gate.

    Example:
        .. code-block:: python

            SqrtXdg * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "SqrtXdg"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply SXdg gate.

        Example:
            .. code-block:: python

                SqrtXdg * qr[0]

        Raises:
            TypeError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1 - 1j, 1 + 1j], [1 + 1j, 1 - 1j]]) / 2

    def inv(self):
        """Apply inverse gate."""
        gate = SqrtXdgGate()
        gate.is_inverse = not self.is_inverse
        return gate


SqrtXdg = SqrtXdgGate()



from .bit import Bit
from .classical_bit import CBit
from .classical_reg import CReg
from .command import Command
from .counter import Counter
from .qubit import QuBit
from .qureg import Qureg
from .circuit import QCircuit
from .parameter import Parameter
